@@ -94,13 +94,26 @@ def findCalibration(fileName):
     calibration = yearCalib[calibrationName]
     return calibration,calibrationName
 
+# Custom class designed to perform a lazy q calculation. Usage:
+# All calculations are needed: q[None]
+# Only a specific slice is needed: q[:10]
+class lazyQ(object):
+    def __init__(self,rotationMatrix,q_temp):
+        self.rotationMatrix = rotationMatrix
+        self.q_temp = q_temp
+        
+    def __getitem__(self,sl=None):
+        return np.einsum('jki,k...->ji...',self.rotationMatrix[:,:,sl].reshape(3,3,-1),self.q_temp)
 
+
+HDFCounts = 'entry/DMC/detector/data'
 ## Dictionary for holding hdf position of attributes. HDFTranslation['a3'] gives hdf position of 'a3'
 HDFTranslation = {'sample':'/entry/sample',
                   'sampleName':'/entry/sample/name',
-                  'monitor':'entry/monitor/monitor',
+                  'monitor':None,#'entry/monitor/monitor',
+                  'monitor1':'entry/monitor/monitor1',
                   'unitCell':'/entry/sample/unit_cell',
-                  'counts':'entry/DMC/detector/data',
+                  #'counts':'entry/DMC/detector/data',
                   'summedCounts': 'entry/DMC/detector/summed_counts',
                   'monochromatorCurvature':'entry/DMC/monochromator/curvature',
                   'monochromatorVerticalCurvature':'entry/DMC/monochromator/curvature_vertical',
@@ -113,11 +126,12 @@ HDFTranslation = {'sample':'/entry/sample',
                   
 
                   'wavelength':'entry/DMC/monochromator/wavelength',
+                  'wavelength_raw':'entry/DMC/monochromator/wavelength_raw',
                   'twoThetaPosition':'entry/DMC/detector/detector_position',
                   'mode':'entry/monitor/mode',
                   'preset':'entry/monitor/preset',
                   'startTime':'entry/start_time',
-                  'time':'entry/monitor/time',
+                  'time':'Henning',# Is to be caught by HDFTranslationAlternatives 'entry/monitor/time',
                   'endTime':'entry/end_time',
                   'comment':'entry/comment',
                   'proposal':'entry/proposal_id',
@@ -130,13 +144,20 @@ HDFTranslation = {'sample':'/entry/sample',
                   'address':'entry/user/address',
                   'affiliation':'entry/user/affiliation',
                   'A3':'entry/sample/rotation_angle',
+                  'se_r':'entry/sample/se_r',
                   'temperature':'entry/sample/temperature',
                   'magneticField':'entry/sample/magnetic_field',
                   'electricField':'entry/sample/electric_field',
                   'scanCommand':'entry/scancommand',
                   'title':'entry/title',
                   'absoluteTime':'entry/control/absolute_time',
-                  'protonBeam':'entry/proton_beam/data'
+                  'protonBeam':None# 'entry/proton_beam/data'
+}
+
+HDFTranslationAlternatives = { # Alternatives to the above list. NOTTICE: The above positions are not checked if an entry in HDFTranslationAlternatives is present
+    'time':['entry/monitor/time','entry/monitor/monitor'],
+    'monitor':['entry/monitor/monitor','entry/monitor/monitor2'],
+    'protonBeam':['entry/proton_beam/data','entry/monitor/proton_charge']
 }
 
 ## Dictionary for holding standard values 
@@ -152,12 +173,16 @@ HDFTranslationDefault = {'twoThetaPosition':np.array([0.0]),
                          'affiliation': 'Unknown Affiliation',
                          'scanCommand': 'Unknown scanCommand',
 
+                         'wavelength_raw':np.array([2.0]),
+                         'monitor1':np.array([0.0]),
+
                          'temperature': np.array([0.0]),
                          'magneticField': np.array([0.0]),
                          'electricField': np.array([0.0]),
 
                          'absoluteTime': np.array([0.0]),
                          'protonBeam': np.array([0.0]),
+                         'se_r': np.array([0.0]),
                          
                          
 
@@ -170,11 +195,12 @@ HDFTranslationFunctions['sampleName'] = [['__getitem__',[0]],['decode',['utf8']]
 HDFTranslationFunctions['mode'] = [['__getitem__',[0]],['decode',['utf8']]]
 HDFTranslationFunctions['startTime'] = [['__getitem__',[0]],['decode',['utf8']]]
 HDFTranslationFunctions['wavelength'] = [['mean',[]]]
+HDFTranslationFunctions['wavelength_raw'] = [['mean',[]]]
 HDFTranslationFunctions['twoThetaPosition'] = [['__getitem__',[0]]]
 HDFTranslationFunctions['endTime'] = [['__getitem__',[0]]]
 HDFTranslationFunctions['experimentalIdentifier'] = [['__getitem__',[0]]]
 HDFTranslationFunctions['comment'] = [['__getitem__',[0]],['decode',['utf8']]]
-HDFTranslationFunctions['proposal'] = [['__getitem__',[0]]]
+HDFTranslationFunctions['proposal'] = [['__getitem__',[0]],['decode',['utf8']]]
 HDFTranslationFunctions['proposalTitle'] = [['__getitem__',[0]],['decode',['utf8']]]
 HDFTranslationFunctions['localContact'] = [['__getitem__',[0]],['decode',['utf8']]]
 HDFTranslationFunctions['proposalUser'] = [['__getitem__',[0]],['decode',['utf8']]]
@@ -195,6 +221,7 @@ HDFInstrumentTranslationFunctions = defaultdict(lambda : [])
 # HDFInstrumentTranslationFunctions['counts'] = [['swapaxes',[1,2]]]
 HDFInstrumentTranslationFunctions['twoThetaPosition'] = [['mean',]]
 HDFInstrumentTranslationFunctions['wavelength'] = [['mean',]]
+HDFInstrumentTranslationFunctions['wavelength_raw'] = [['mean',]]
 
 extraAttributes = ['name','fileLocation']
 
@@ -203,6 +230,7 @@ possibleAttributes.sort(key=lambda v: v.lower())
 
 HDFTypes = defaultdict(lambda: lambda x: np.array([np.string_(x)]))
 HDFTypes['monitor'] = np.array
+HDFTypes['monitor1'] = np.array
 HDFTypes['monochromatorCurvature'] = np.array
 HDFTypes['monochromatorVerticalCurvature'] = np.array
 HDFTypes['monochromatorGoniometerLower'] = np.array
@@ -212,6 +240,7 @@ HDFTypes['monochromatorTakeoffAngle'] = np.array
 HDFTypes['monochromatorTranslationLower'] = np.array
 HDFTypes['monochromatorTranslationUpper'] = np.array
 HDFTypes['wavelength'] = np.array
+HDFTypes['wavelength_raw'] = np.array
 HDFTypes['twoThetaPosition'] = np.array
 # HDFTypes['mode'] = lambda x: np.array([np.string_(x)])
 HDFTypes['preset'] = np.array
@@ -235,7 +264,10 @@ HDFUnits = {
     'monochromatorTranslationUpper':'mm',
     'twoThetaPosition':'degree',
     'monitor':'counts',
-    'wavelength':'A'
+    'monitor1':'counts',
+    'protonBeam':'uA',
+    'wavelength':'A',
+    'wavelength_raw':'A'
 }
 
 
@@ -274,13 +306,15 @@ def loadDataFile(fileLocation=None,fileType='Unknown',**kwargs):
         raise FileNotFoundError('Provided file path "{}" not found.'.format(fileLocation))
 
     A3 = shallowRead([fileLocation],['A3'])[0]['A3']
+
+    se_r = shallowRead([fileLocation],['se_r'])[0]['se_r']
     
     T = 'Unknown' # type of datafile
 
     if A3 is None: # there is no A3 values at all
         T = 'powder'
         
-    elif len(A3) == 1:
+    elif len(A3) == 1 and se_r.any() is None:
         T = 'powder'
     else:
         T = 'singlecrystal'
@@ -295,7 +329,7 @@ def loadDataFile(fileLocation=None,fileType='Unknown',**kwargs):
         df = DataFile(fileLocation)
 
     
-    repeats = df.counts.shape[1]
+    repeats = df.countShape[1]
     # Insert standard values if not present in kwargs
     if not 'radius' in kwargs:
         kwargs['radius'] = 0.8
@@ -309,7 +343,7 @@ def loadDataFile(fileLocation=None,fileType='Unknown',**kwargs):
         
     if 'twoThetaPosition' in kwargs:
         if not 'twoTheta' in kwargs:
-            df.twoTheta = np.linspace(0,132,9*128)+df.twoThetaPosition
+            df.twoTheta = np.linspace(0,-132,9*128)+df.twoThetaPosition
         else:
             df.twoTheta = kwargs['twoTheta']
     elif 'twoTheta' in kwargs:
@@ -318,6 +352,10 @@ def loadDataFile(fileLocation=None,fileType='Unknown',**kwargs):
     df.initializeQ()
     df.loadNormalization()
 
+    year,month,date = [int(x) for x in df.startTime.split(' ')[0].split('-')]
+    if year == 2022:
+        df.mask[0,-2,:] = True
+
     return df
 
 
@@ -327,6 +365,7 @@ class DataFile(object):
     def __init__(self, file=None):
         self.fileType = 'DataFile'
         self._twoThetaOffset = 0.0
+        self._counts = None
 
         if not file is None: 
             if isinstance(file,DataFile): # Copy everything from provided file
@@ -340,7 +379,6 @@ class DataFile(object):
             else:
                 raise FileNotFoundError('Provided file path "{}" not found.'.format(file))
 
-    
     @_tools.KwargChecker()
     def loadFile(self,filePath):
         if not os.path.exists(filePath):
@@ -357,14 +395,26 @@ class DataFile(object):
         with hdf.File(filePath,mode='r') as f:
 
             self.sample = Sample.Sample(sample=f.get(HDFTranslation['sample']))
-
+            self.countShape = f.get(HDFCounts).shape
             # load standard things using the shallow read
             instr = getInstrument(f)
+
+            if not f.get('/entry/reduction') is None: # Data file is a merged/reduced data file
+                red = f['/entry/reduction']
+
+                # Complicated way to avoid having to guess the name of the reduction algorithm.....
+                self.original_files = np.asarray([name.decode('UTF8') for name in list(red.values())[0].get('rawdata')]) 
+                
             for parameter in HDFTranslation.keys():
                 if parameter in ['unitCell','sample','unitCell']:
                     continue
-                
-                if parameter in HDFTranslation:
+                if parameter in HDFTranslationAlternatives:
+                    for entry in HDFTranslationAlternatives[parameter]:
+                        value = np.array(f.get(entry))
+                        if not value.shape == ():
+                            break
+
+                elif parameter in HDFTranslation:
                     value = np.array(f.get(HDFTranslation[parameter]))
                     TrF= HDFTranslationFunctions
                 elif parameter in HDFInstrumentTranslation:
@@ -380,8 +430,8 @@ class DataFile(object):
                 
                 setattr(self,parameter,value)
                 
-        self.counts.shape = (1,*self.counts.shape) # Standard shape
-
+        self.countShape = (1,*self.countShape) # Standard shape
+    
     def initializeQ(self):
         if len(self.twoTheta.shape) == 2:
             self.twoTheta, z = np.meshgrid(self.twoTheta[0].flatten(),self.verticalPosition,indexing='xy')
@@ -390,13 +440,13 @@ class DataFile(object):
             
         self.pixelPosition = np.array([self.radius*np.cos(np.deg2rad(self.twoTheta)),
                                     -self.radius*np.sin(np.deg2rad(self.twoTheta)),
-                                    z]).reshape(3,*self.counts.shape[1:])
+                                    z]).reshape(3,*self.countShape[1:])
         
         
         #self.Monitor = self.monitor
         
         if np.any(np.isclose(self.monitor,0)): # error mode from commissioning
-            self.monitor = np.ones(self.counts.shape[0])
+            self.monitor = np.ones(self.countShape[0])
         
         self.alpha = np.rad2deg(np.arctan2(self.pixelPosition[2],self.radius))
         # Above line makes an implicit call to the self.calculateQ method!
@@ -409,26 +459,32 @@ class DataFile(object):
     def loadNormalization(self):
         # Load calibration
         try:
-            self.normalization, self.normalizationFile = findCalibration(self.fileName)
+            if hasattr(self,'original_files'): # We are working with a converted/merged file
+                name = self.original_files[0]
+            else:
+                name = self.fileName
+            self.normalization, self.normalizationFile = findCalibration(name)
         except ValueError:
             self.normalizationFile = 'None'
 
         if self.normalizationFile == 'None':
-            self.normalization = np.ones_like(self.counts,dtype=float)
+            self.normalization = np.ones(self.countShape,dtype=float)
         else:
             
             if self.fileType.lower() == "singlecrystal": # A3 scan
-                self.normalization = np.repeat(self.normalization,self.counts.shape[0],axis=0)
-                #self.normalization.shape = self.counts.shape
-                self.normalization = self.normalization.reshape(self.counts.shape)
+                self.normalization = self.normalization#np.repeat(self.normalization[np.newaxis],self.countShape[0],axis=0)
+                #self.normalization.shape = self.countShape
+                #self.normalization = self.normalization.reshape(self.countShape)
             else:
-                self.normalization = self.normalization.reshape(self.counts.shape)
+                self.normalization = self.normalization.reshape(self.countShape)
 
     def __len__(self):
-        if hasattr(self,'counts'):
-            return len(self.counts)
-        else:
-            return 0
+        if not hasattr(self,'countShape'):
+            if hasattr(self,'counts'):
+                self.countShape = self.counts.shape[0]
+            else:
+                return 0
+        return self.countShape[0]
 
     @property
     def A3(self):
@@ -469,7 +525,7 @@ class DataFile(object):
             self._detector_position = np.array([0.0]*len(self.A3))
         else:
             self._detector_position = np.asarray(twoTheta)
-        self.twoTheta = np.repeat((np.linspace(0,132,1152) + self._detector_position + self._twoThetaOffset)[np.newaxis],self.counts.shape[1],axis=0)
+        self.twoTheta = np.repeat((np.linspace(0,-132,1152) + self._detector_position + self._twoThetaOffset)[np.newaxis],self.countShape[1],axis=0)
         if hasattr(self,'_Ki') and hasattr(self,'twoTheta'):
             self.calculateQ()
 
@@ -500,7 +556,7 @@ class DataFile(object):
     @twoThetaOffset.setter
     def twoThetaOffset(self,dTheta):
         self._twoThetaOffset = dTheta
-        self.twoTheta = np.repeat((np.linspace(0,132,1152) + self._detector_position + self._twoThetaOffset)[np.newaxis],self.counts.shape[1],axis=0)
+        self.twoTheta = np.repeat((np.linspace(0,132,1152) + self._detector_position + self._twoThetaOffset)[np.newaxis],self.countShape[1],axis=0)
         self.calculateQ()
 
     @property
@@ -518,7 +574,7 @@ class DataFile(object):
         self._wavelength = wavelength
         self._Ki = 2*np.pi/wavelength
         self.calculateQ()
-
+    
     def calculateQ(self):
         """Calculate Q and qx,qy,qz using the current A3 values"""
         if not (hasattr(self,'Ki') and hasattr(self,'twoTheta')
@@ -530,30 +586,34 @@ class DataFile(object):
         self.kf = self.Ki * np.array([-np.sin(np.deg2rad(self.twoTheta))*np.cos(np.deg2rad(self.alpha)),
                                     np.cos(np.deg2rad(self.twoTheta))*np.cos(np.deg2rad(self.alpha)),
                                     np.sin(np.deg2rad(self.alpha))])
-        self.q = self.ki-self.kf   
+           
         if self.fileType.lower() == 'singlecrystal': # A3 Scan
             # rotate kf to correct for A3
             zero = np.zeros_like(self.A3)
             ones = np.ones_like(self.A3)
-            rotMat = np.array([[np.cos(np.deg2rad(self.A3)),np.sin(np.deg2rad(self.A3)),zero],[-np.sin(np.deg2rad(self.A3)),np.cos(np.deg2rad(self.A3)),zero],[zero,zero,ones]])
-            q_temp = self.ki-self.kf
+            self.rotMat = np.array([[np.cos(np.deg2rad(self.A3)),np.sin(np.deg2rad(self.A3)),zero],[-np.sin(np.deg2rad(self.A3)),np.cos(np.deg2rad(self.A3)),zero],[zero,zero,ones]])
+            self.q_temp = self.ki-self.kf
 
-            self.q = np.einsum('jki,k...->ji...',rotMat,q_temp)
-            self.Q = np.linalg.norm(self.q,axis=0)
+            self.q = lazyQ(self.rotMat, self.q_temp)
+
+            self.Q = np.linalg.norm(self.q[None],axis=0)
         else:
-            self.Q = np.array([np.linalg.norm(self.q,axis=0)])
+            self.qLocal = self.ki-self.kf
+            self.Q = np.array([np.linalg.norm(self.qLocal,axis=0)])
 
-        self.correctedTwoTheta = 2.0*np.rad2deg(np.arcsin(self.wavelength*self.Q[0]/(4*np.pi)))[np.newaxis].repeat(self.Q.shape[0],axis=0)
-        self.phi = np.rad2deg(np.arctan2(self.q[2],np.linalg.norm(self.q[:2],axis=0)))
+        #self.correctedTwoTheta = 2.0*np.rad2deg(np.arcsin(self.wavelength*self.Q[0]/(4*np.pi)))[np.newaxis].repeat(self.Q.shape[0],axis=0)
+        
         
 
-    def generateMask(self,maskingFunction = maskFunction, **pars):
+    def generateMask(self,maskingFunction = maskFunction, replace=True, **pars):
         """Generate mask to applied to data in data file
         
         Kwargs:
 
             - maskingFunction (function): Function called on self.phi to generate mask (default maskFunction)
 
+            - replace (bool): If true new mask replaces old one, otherwise add together (default True)
+
         All other arguments are passed to the masking function.
 
         """
@@ -564,9 +624,15 @@ class DataFile(object):
             raise RuntimeError('DataFile does not contain any counts. Look for self.counts but found nothing.')
 
         if maskingFunction is None:
-            self.mask = np.zeros_like(self.counts,dtype=bool)
+            if replace:
+                self.mask = np.zeros(self.countShape,dtype=bool)
+            else:
+                self.mask += np.zeros(self.countShape,dtype=bool)
         else:
-            self.mask = maskingFunction(self.phi,**pars)
+            if replace:
+                self.mask = maskingFunction(self.phi,**pars).reshape(*self.countShape)
+            else:
+                self.mask += maskingFunction(self.phi,**pars).reshape(*self.countShape)
         
         
 
@@ -585,14 +651,14 @@ class DataFile(object):
 
 
     @_tools.KwargChecker(function=plt.errorbar,include=_tools.MPLKwargs)
-    def plotDetector(self,ax=None,applyNormalization=True,**kwargs):
+    def plotDetector(self,ax=None,applyCalibration=True,**kwargs):
         """Plot intensity as function of twoTheta (and vertical position of pixel in 2D)
 
         Kwargs:
 
             - ax (axis): Matplotlib axis into which data is plotted (default None - generates new)
 
-            - applyNormalization (bool): If true, take detector efficiency into account (default True)
+            - applyCalibration (bool): If true, take detector efficiency into account (default True)
 
             - All other key word arguments are passed on to plotting routine
 
@@ -604,14 +670,15 @@ class DataFile(object):
             fig = ax.get_figure()
 
         
-        intensity = self.counts/self.monitor.reshape(-1,1,1)
-        if applyNormalization:
+        if applyCalibration:
             intensity=self.intensity/self.monitor.reshape(-1,1,1)#1.0/self.normalization
 
         count_err = np.sqrt(self.counts)
         intensity_err = count_err/self.monitor.reshape(-1,1,1)
-        if applyNormalization:
-            intensity_err*=1.0/self.normalization
+        if applyCalibration:
+            with warnings.catch_warnings():
+                warnings.simplefilter("ignore")
+                intensity_err*=1.0/self.normalization
  
 
 
@@ -812,27 +879,67 @@ class DataFile(object):
         return dif
 
 
+    @property
+    def counts(self):
+        if self._counts is None:
+            with hdf.File(os.path.join(self.folder,self.fileName),mode='r') as f:
+                return np.array(f.get(HDFCounts)).reshape(self.countShape)
+        else:
+            return self._counts.reshape(self.countShape)
+    
+    def countsSliced(self,sl):
+        if self._counts is None:
+            with hdf.File(os.path.join(self.folder,self.fileName),mode='r') as f:
+                return np.array(f.get(HDFCounts)[sl])
+        else:
+            return self._counts[sl]
+            
+
     @property
     def intensity(self):
-        return np.divide(self.counts,self.normalization)
+        with warnings.catch_warnings():
+            warnings.simplefilter("ignore")
+            if self.fileType.lower() == 'singlecrystal':
+                return np.divide(self.counts,self.normalization[np.newaxis])
+            else:
+                return np.divide(self.counts,self.normalization)
+
+    
+    def intensitySliced(self,sl):
+        with warnings.catch_warnings():
+            warnings.simplefilter("ignore")
+            if self.fileType.lower() == 'singlecrystal':
+                return np.divide(self.countsSliced(sl),self.normalization[np.newaxis])
+            else:
+                return np.divide(self.countsSliced(sl),self.normalization)
+            #return np.divide(self.countsSliced(sl),self.normalization[sl])
+
+    
 
     def InteractiveViewer(self,**kwargs):
         if not self.fileType.lower() in ['singlecrystal','powder'] :
             raise AttributeError('Interactive Viewer can only be used for the new data files. Either for powder or for a single crystal A3 scan')
         return InteractiveViewer.InteractiveViewer(self.intensity,self.twoTheta,self.pixelPosition,self.A3,scanParameter = 'A3',scanValueUnit='deg',colorbar=True,**kwargs)
 
+    @property
+    def correctedTwoTheta(self):
+        return 2.0*np.rad2deg(np.arcsin(self.wavelength*self.Q[0]/(4*np.pi)))[np.newaxis].repeat(self.Q.shape[0],axis=0)
 
+    
 
-
-
-
+    @property
+    def phi(self):
+        if self.fileType.lower() == 'singlecrystal':
+            return np.rad2deg(np.arctan2(self.q[None][2],np.linalg.norm(self.q[None][:2],axis=0)))
+        else:
+            return np.rad2deg(np.arctan2(self.qLocal[2],np.linalg.norm(self.qLocal[:2],axis=0)))
 
 
 class SingleCrystalDataFile(DataFile):
     def __init__(self,fileType):
         super(SingleCrystalDataFile,self).__init__(fileType)
         self.fileType = 'SingleCrystal'
-        self.counts.shape = (-1,128,1152)
+        self.countShape = (self.countShape[0]*self.countShape[1],128,1152)
 
 class PowderDataFile(DataFile):
     def __init__(self,fileType):
@@ -871,6 +978,13 @@ def shallowRead(files,parameters):
                     v = os.path.dirname(file)
                     vals[p] = v
                     continue
+                elif p in HDFTranslationAlternatives:
+                    for entry in HDFTranslationAlternatives[p]:
+                        v = np.array(f.get(entry))
+                        if not v.shape == ():
+                            TrF= HDFTranslationFunctions
+                            break
+
                 elif p in HDFTranslation:
                     v = np.array(f.get(HDFTranslation[p]))
                     TrF= HDFTranslationFunctions
