@@ -186,7 +186,11 @@ class Viewer3D(object):
             maxVal = np.nanmax(self.masked_array[np.isfinite(self.masked_array)])
         except ValueError:
             maxVal = 1
-        self.caxis = [np.nanmin(self.masked_array),maxVal]
+        try:
+            minVal = np.nanmin(self.masked_array[np.isfinite(self.masked_array)])
+        except ValueError:
+            minVal = 0
+        self.caxis = [minVal,maxVal]
         if self.grid:
             self.ax.grid(True,zorder=self.gridZOrder)
         else:
